diff src/redisplay.c @ 853:2b6fa2618f76

[xemacs-hg @ 2002-05-28 08:44:22 by ben] merge my stderr-proc ws make-docfile.c: Fix places where we forget to check for EOF. code-init.el: Don't use CRLF conversion by default on process output. CMD.EXE and friends work both ways but Cygwin programs don't like the CRs. code-process.el, multicast.el, process.el: Removed. Improvements to call-process-internal: -- allows a buffer to be specified for input and stderr output -- use it on all systems -- implement C-g as documented -- clean up and comment call-process-region uses new call-process facilities; no temp file. remove duplicate funs in process.el. comment exactly how coding systems work and fix various problems. open-multicast-group now does similar coding-system frobbing to open-network-stream. dumped-lisp.el, faces.el, msw-faces.el: Fix some hidden errors due to code not being defined at the right time. xemacs.mak: Add -DSTRICT. ================================================================ ALLOW SEPARATION OF STDOUT AND STDERR IN PROCESSES ================================================================ Standard output and standard error can be processed separately in a process. Each can have its own buffer, its own mark in that buffer, and its filter function. You can specify a separate buffer for stderr in `start-process' to get things started, or use the new primitives: set-process-stderr-buffer process-stderr-buffer process-stderr-mark set-process-stderr-filter process-stderr-filter Also, process-send-region takes a 4th optional arg, a buffer. Currently always uses a pipe() under Unix to read the error output. (#### Would a PTY be better?) sysdep.h, sysproc.h, unexfreebsd.c, unexsunos4.c, nt.c, emacs.c, callproc.c, symsinit.h, sysdep.c, Makefile.in.in, process-unix.c: Delete callproc.c. Move child_setup() to process-unix.c. wait_for_termination() now only needed on a few really old systems. console-msw.h, event-Xt.c, event-msw.c, event-stream.c, event-tty.c, event-unixoid.c, events.h, process-nt.c, process-unix.c, process.c, process.h, procimpl.h: Rewrite the process methods to handle a separate channel for error input. Create Lstreams for reading in the error channel. Many process methods need change. In general the changes are fairly clear as they involve duplicating what's used for reading the normal stdout and changing for stderr -- although tedious, as such changes are required throughout the entire process code. Rewrote the code that reads process output to do two loops, one for stdout and one for stderr. gpmevent.c, tooltalk.c: set_process_filter takes an argument for stderr. ================================================================ NEW ERROR-TRAPPING MECHANISM ================================================================ Totally rewrite error trapping code to be unified and support more features. Basic function is call_trapping_problems(), which lets you specify, by means of flags, what sorts of problems you want trapped. these can include -- quit -- errors -- throws past the function -- creation of "display objects" (e.g. buffers) -- deletion of already-existing "display objects" (e.g. buffers) -- modification of already-existing buffers -- entering the debugger -- gc -- errors->warnings (ala suspended errors) etc. All other error funs rewritten in terms of this one. Various older mechanisms removed or rewritten. window.c, insdel.c, console.c, buffer.c, device.c, frame.c: When creating a display object, added call to note_object_created(), for use with trapping_problems mechanism. When deleting, call check_allowed_operation() and note_object deleted(). The trapping-problems code records the objects created since the call-trapping-problems began. Those objects can be deleted, but none others (i.e. previously existing ones). bytecode.c, cmdloop.c: internal_catch takes another arg. eval.c: Add long comments describing the "five lists" used to maintain state (backtrace, gcpro, specbind, etc.) in the Lisp engine. backtrace.h, eval.c: Implement trapping-problems mechanism, eliminate old mechanisms or redo in terms of new one. frame.c, gutter.c: Flush out the concept of "critical display section", defined by the in_display() var. Use an internal_bind() to get it reset, rather than just doing it at end, because there may be a non-local exit. event-msw.c, event-stream.c, console-msw.h, device.c, dialog-msw.c, frame.c, frame.h, intl.c, toolbar.c, menubar-msw.c, redisplay.c, alloc.c, menubar-x.c: Make use of new trapping-errors stuff and rewrite code based on old mechanisms. glyphs-widget.c, redisplay.h: Protect calling Lisp in redisplay. insdel.c: Protect hooks against deleting existing buffers. frame-msw.c: Use EQ, not EQUAL in hash tables whose keys are just numbers. Otherwise we run into stickiness in redisplay because internal_equal() can QUIT. ================================================================ SIGNAL, C-G CHANGES ================================================================ Here we change the way that C-g interacts with event reading. The idea is that a C-g occurring while we're reading a user event should be read as C-g, but elsewhere should be a QUIT. The former code did all sorts of bizarreness -- requiring that no QUIT occurs anywhere in event-reading code (impossible to enforce given the stuff called or Lisp code invoked), and having some weird system involving enqueue/dequeue of a C-g and interaction with Vquit_flag -- and it didn't work. Now, we simply enclose all code where we want C-g read as an event with {begin/end}_dont_check_for_quit(). This completely turns off the mechanism that checks (and may remove or alter) C-g in the read-ahead queues, so we just get the C-g normal. Signal.c documents this very carefully. cmdloop.c: Correct use of dont_check_for_quit to new scheme, remove old out-of-date comments. event-stream.c: Fix C-g handling to actually work. device-x.c: Disable quit checking when err out. signal.c: Cleanup. Add large descriptive comment. process-unix.c, process-nt.c, sysdep.c: Use QUIT instead of REALLY_QUIT. It's not necessary to use REALLY_QUIT and just confuses the issue. lisp.h: Comment quit handlers. ================================================================ CONS CHANGES ================================================================ free_cons() now takes a Lisp_Object not the result of XCONS(). car and cdr have been renamed so that they don't get used directly; go through XCAR(), XCDR() instead. alloc.c, dired.c, editfns.c, emodules.c, fns.c, glyphs-msw.c, glyphs-x.c, glyphs.c, keymap.c, minibuf.c, search.c, eval.c, lread.c, lisp.h: Correct free_cons calling convention: now takes Lisp_Object, not Lisp_Cons chartab.c: Eliminate direct use of ->car, ->cdr, should be black box. callint.c: Rewrote using EXTERNAL_LIST_LOOP to avoid use of Lisp_Cons. ================================================================ USE INTERNAL-BIND-* ================================================================ eval.c: Cleanups of these funs. alloc.c, fileio.c, undo.c, specifier.c, text.c, profile.c, lread.c, redisplay.c, menubar-x.c, macros.c: Rewrote to use internal_bind_int() and internal_bind_lisp_object() in place of whatever varied and cumbersome mechanisms were formerly there. ================================================================ SPECBIND SANITY ================================================================ backtrace.h: - Improved comments backtrace.h, bytecode.c, eval.c: Add new mechanism check_specbind_stack_sanity() for sanity checking code each time the catchlist or specbind stack change. Removed older prototype of same mechanism. ================================================================ MISC ================================================================ lisp.h, insdel.c, window.c, device.c, console.c, buffer.c: Fleshed out authorship. device-msw.c: Correct bad Unicode-ization. print.c: Be more careful when not initialized or in fatal error handling. search.c: Eliminate running_asynch_code, an FSF holdover. alloc.c: Added comments about gc-cons-threshold. dialog-x.c: Use begin_gc_forbidden() around code to build up a widget value tree, like in menubar-x.c. gui.c: Use Qunbound not Qnil as the default for gethash. lisp-disunion.h, lisp-union.h: Added warnings on use of VOID_TO_LISP(). lisp.h: Use ERROR_CHECK_STRUCTURES to turn on ERROR_CHECK_TRAPPING_PROBLEMS and ERROR_CHECK_TYPECHECK lisp.h: Add assert_with_message. lisp.h: Add macros for gcproing entire arrays. (You could do this before but it required manual twiddling the gcpro structure.) lisp.h: Add prototypes for new functions defined elsewhere.
author ben
date Tue, 28 May 2002 08:45:36 +0000
parents e7ee5f8bde58
children 804517e16990
line wrap: on
line diff
--- a/src/redisplay.c	Sat May 25 01:55:30 2002 +0000
+++ b/src/redisplay.c	Tue May 28 08:45:36 2002 +0000
@@ -33,11 +33,24 @@
 /*****************************************************************************
  The Golden Rules of Redisplay
 
- First:		It Is Better To Be Correct Than Fast
- Second:	Thou Shalt Not Run Elisp From Within Redisplay
- Third:		It Is Better To Be Fast Than Not To Be
+ First:	  It Is Better To Be Correct Than Fast
+ Second:  Thou Shalt Use Due Caution When Running Elisp From Within Redisplay
+ Third:   It Is Better To Be Fast Than Not To Be
  ****************************************************************************/
 
+/* Note: The second rule used to prohibit running Elisp from within redisplay,
+   but that's not correct.
+
+   Use
+
+   callN_trapping_problems (..., INHIBIT_GC
+                            | INHIBIT_ANY_CHANGE_AFFECTING_REDISPLAY)
+
+   instead.
+
+   --ben
+*/
+
 #include <config.h>
 #include "lisp.h"
 
@@ -46,6 +59,7 @@
 #include "debug.h"
 #include "device.h"
 #include "elhash.h"
+#include "events.h"
 #include "extents.h"
 #include "faces.h"
 #include "frame.h"
@@ -6486,10 +6500,13 @@
 	{
 	  Lisp_Object window = wrap_window (w);
 
-	  va_run_hook_with_args_in_buffer (XBUFFER (w->buffer),
-					   Qredisplay_end_trigger_functions,
-					   2, window,
-					   w->redisplay_end_trigger);
+	  va_run_hook_with_args_in_buffer_trapping_problems
+	    ("Error in redisplay end trigger",
+	     XBUFFER (w->buffer),
+	     Qredisplay_end_trigger_functions,
+	     2, window,
+	     w->redisplay_end_trigger,
+	     INHIBIT_EXISTING_PERMANENT_DISPLAY_OBJECT_DELETION);
 	  w->redisplay_end_trigger = Qnil;
 	}
     }
@@ -6503,6 +6520,7 @@
 redisplay_frame (struct frame *f, int preemption_check)
 {
   struct device *d = XDEVICE (f->device);
+  int depth;
 
   if (preemption_check
       && !DEVICE_IMPL_FLAG (d, XDEVIMPF_DONT_PREEMPT_REDISPLAY))
@@ -6541,7 +6559,7 @@
     }
 
   /* The menubar, toolbar, and icon updates must be done before
-     hold_frame_size_changes is called and we are officially
+     enter_redisplay_critical_section is called and we are officially
      'in_display'.  They may eval lisp code which may call Fsignal.
      If in_display is set Fsignal will abort. */
 
@@ -6580,7 +6598,7 @@
       reset_gutter_display_lines (f);
     }
 
-  hold_frame_size_changes ();
+  depth = enter_redisplay_critical_section ();
 
   /* ----------------- BEGIN CRITICAL REDISPLAY SECTION ---------------- */
   /* Within this section, we are defenseless and assume that the
@@ -6590,7 +6608,7 @@
      2) Lisp code evaluation
      3) frame size changes
 
-     We ensure (3) by calling hold_frame_size_changes(), which
+     We ensure (3) by calling enter_redisplay_critical_section(), which
      will cause any pending frame size changes to get put on hold
      till after the end of the critical section.  (1) follows
      automatically if (2) is met.  #### Unfortunately, there are
@@ -6644,10 +6662,8 @@
 
   /* ----------------- END CRITICAL REDISPLAY SECTION ---------------- */
 
-  /* Allow frame size changes to occur again.
-
-     #### what happens if changes to other frames happen? */
-  unhold_one_frame_size_changes (f);
+  /* Allow frame size changes to occur again. */
+  exit_redisplay_critical_section (depth);
 
   map_windows (f, call_redisplay_end_triggers, 0);
   return 0;
@@ -6750,13 +6766,6 @@
   return 0;
 }
 
-static Lisp_Object
-restore_profiling_redisplay_flag (Lisp_Object val)
-{
-  profiling_redisplay_flag = XINT (val);
-  return Qnil;
-}
-
 /* Ensure that all windows on all frames on all devices are displaying
    the current contents of their respective buffers. */
 
@@ -6768,11 +6777,7 @@
   int count = specpdl_depth ();
 
   if (profiling_active)
-    {
-      record_unwind_protect (restore_profiling_redisplay_flag,
-			     make_int (profiling_redisplay_flag));
-      profiling_redisplay_flag = 1;
-    }
+    internal_bind_int (&profiling_redisplay_flag, 1);
 
   if (asynch_device_change_pending)
     handle_asynch_device_change ();
@@ -6826,36 +6831,72 @@
 #endif /* ERROR_CHECK_DISPLAY */
 }
 
+/* Note: All places in the C code that call redisplay() are prepared
+   to handle GCing.  However, we can't currently handle GC inside the
+   guts of redisplay (#### someone should fix this), so we need to use
+   INHIBIT_GC when calling Lisp.
+
+   #### We probably can't handle any deletion of existing buffers, frames,
+   windows, devices, consoles, text changes, etc. either.  We should
+
+   (a) Create the appropriate INHIBIT_ flags for this.
+   (b) In the longer run, fix redisplay to handle this.
+
+   (#### What about other external entry points to the redisplay code?
+   Someone should go through and make sure that all callers can handle
+   GC there, too.)
+*/
+
 void
 redisplay (void)
 {
+  run_pre_idle_hook ();
+  redisplay_no_pre_idle_hook ();
+}
+
+void
+redisplay_no_pre_idle_hook (void)
+{
   if (last_display_warning_tick != display_warning_tick &&
       !inhibit_warning_display)
     {
       /* If an error occurs during this function, oh well.
          If we report another warning, we could get stuck in an
 	 infinite loop reporting warnings. */
-      call0_trapping_errors (0, Qdisplay_warning_buffer);
+      call0_trapping_problems
+	(0, Qdisplay_warning_buffer,
+	 INHIBIT_EXISTING_PERMANENT_DISPLAY_OBJECT_DELETION);
       last_display_warning_tick = display_warning_tick;
     }
   /* The run_hook_trapping_errors functions are smart enough not
      to do any evalling if the hook function is empty, so there
-     should not be any significant time loss.  All places in the
-     C code that call redisplay() are prepared to handle GCing,
-     so we should be OK. */
+     should not be any significant time loss. */
 #ifndef INHIBIT_REDISPLAY_HOOKS
-  run_hook_trapping_errors ("Error in pre-redisplay-hook",
-			    Qpre_redisplay_hook);
+  run_hook_trapping_problems
+    ("Error in pre-redisplay-hook",
+     Qpre_redisplay_hook,
+     INHIBIT_EXISTING_PERMANENT_DISPLAY_OBJECT_DELETION);
 #endif /* INHIBIT_REDISPLAY_HOOKS */
 
   redisplay_without_hooks ();
 
 #ifndef INHIBIT_REDISPLAY_HOOKS
-  run_hook_trapping_errors ("Error in post-redisplay-hook",
-			    Qpost_redisplay_hook);
+  run_hook_trapping_problems
+    ("Error in post-redisplay-hook",
+     Qpost_redisplay_hook,
+     INHIBIT_EXISTING_PERMANENT_DISPLAY_OBJECT_DELETION);
 #endif /* INHIBIT_REDISPLAY_HOOKS */
 }
 
+Lisp_Object
+eval_within_redisplay (Lisp_Object dont_trust_this_damn_sucker)
+{
+  return
+    eval_in_buffer_trapping_problems
+    ("Error calling function within redisplay", current_buffer,
+     dont_trust_this_damn_sucker,
+     INHIBIT_GC | INHIBIT_ANY_CHANGE_AFFECTING_REDISPLAY);
+}
 
 /* Efficiently determine the window line number, and return a pointer
    to its printed representation.  Do this regardless of whether
@@ -9014,6 +9055,7 @@
       DEVICE_FRAME_LOOP (frmcons, d)
 	{
 	  struct frame *f = XFRAME (XCAR (frmcons));
+	  int depth;
 
 	  if (FRAME_REPAINT_P (f) && FRAME_HAS_MINIBUF_P (f))
 	    {
@@ -9032,7 +9074,9 @@
 		  MAYBE_DEVMETH (d, clear_frame, (f));
 		  f->echo_area_garbaged = 0;
 		}
+	      depth = enter_redisplay_critical_section ();
 	      redisplay_window (window, 0);
+	      exit_redisplay_critical_section (depth);
 	      MAYBE_DEVMETH (d, frame_output_end, (f));
 
 	      call_redisplay_end_triggers (XWINDOW (window), 0);
@@ -9043,13 +9087,6 @@
   return Qnil;
 }
 
-static Lisp_Object
-restore_disable_preemption_value (Lisp_Object value)
-{
-  disable_preemption = XINT (value);
-  return Qnil;
-}
-
 DEFUN ("redraw-frame", Fredraw_frame, 0, 2, 0, /*
 Clear frame FRAME and output again what is supposed to appear on it.
 FRAME defaults to the selected frame if omitted.
@@ -9063,11 +9100,7 @@
   int count = specpdl_depth ();
 
   if (!NILP (no_preempt))
-    {
-      record_unwind_protect (restore_disable_preemption_value,
-			     make_int (disable_preemption));
-      disable_preemption++;
-    }
+    internal_bind_int (&disable_preemption, 1 + disable_preemption);
 
   f->clear = 1;
   redisplay_frame (f, 1);
@@ -9094,11 +9127,7 @@
   int count = specpdl_depth ();
 
   if (!NILP (no_preempt))
-    {
-      record_unwind_protect (restore_disable_preemption_value,
-			     make_int (disable_preemption));
-      disable_preemption++;
-    }
+    internal_bind_int (&disable_preemption, 1 + disable_preemption);
 
   redisplay_frame (f, 1);
 
@@ -9128,11 +9157,7 @@
   int count = specpdl_depth ();
 
   if (!NILP (no_preempt))
-    {
-      record_unwind_protect (restore_disable_preemption_value,
-			     make_int (disable_preemption));
-      disable_preemption++;
-    }
+    internal_bind_int (&disable_preemption, 1 + disable_preemption);
 
   DEVICE_FRAME_LOOP (frmcons, d)
     {
@@ -9155,6 +9180,9 @@
 Normally, redisplay is preempted as normal if input arrives.  However,
 if optional second arg NO-PREEMPT is non-nil, redisplay will not stop for
 input and is guaranteed to proceed to completion.
+
+Note: If you simply want everything redisplayed, the current idiom is
+`(sit-for 0)'.
 */
        (device, no_preempt))
 {
@@ -9162,11 +9190,7 @@
   int count = specpdl_depth ();
 
   if (!NILP (no_preempt))
-    {
-      record_unwind_protect (restore_disable_preemption_value,
-			     make_int (disable_preemption));
-      disable_preemption++;
-    }
+    internal_bind_int (&disable_preemption, 1 + disable_preemption);
 
   redisplay_device (d, 0);
@@ -6826,36 +6831,72 @@
       last_display_warning_tick = display_warning_tick;
     }
-  /* The run_hook_trapping_errors functions are smart enough not
+  /* The run_hook_trapping_problems functions are smart enough not
      to do any evalling if the hook function is empty, so there